%% Covariance matrix of parameters of essential matrix from I > 1 ACs
%
% CovM = CovM_E_matrix_AC(x0, z0, A, Caa, B, R, indk)
%                        (U0, V0, A, Caa, b, R, s)
% in the note "Precision of Essential Matrix from Affne Matches"
%
% model
%        Xm' * E * zh + Zm' * E' * x0h = 0
%        with
%        Xm = [P0'|x0h], Zm = [A_0|zh]
%        P0 = [I_2;[0 0]], A0 = [A(1:2,1:2); [0,0]]
%        x0 = [x0;1], z = z0 + A(1:2,3)
%        E = S(B) R'
%        x0 = [I |  0] Xv
%        z  = [R | -B] Xv
%
% x0  = 2 x 2 matrix of image coordiantes of reference points x0(i,:) 1.
% z0  = 2 x 2 matrix of image coordiantes of reference points z0(i,:) 2.
% A   = 6 x 3 matrix of affinities (stochastic)
% Caa = 12 x 6 matrix of Covariance matrices of affinities
% R   = 3 x 3 rotation matrix
% B   = 3 x 1 normalized direction of base line, 1-->2
% indk = 5-vector of indices chosen from rows of C'
%
% end = v: vector, end = m: matrix
%
% Checked in ?
%
%  see: report on affinities to slopes and E
%
% Wolfgang Frstner 2019-11-26
% wfoerstn@uni-bonn.de
% 

function CovM = CovM_bR_from_ACs_5(x0m, z0m, Am, Caam, Bv, Rm, indk)


% essential matrix
Em = calc_S(Bv) * Rm';

% fitted right points z (V)
zm = z0m + [Am(1:2,3)';Am(4:5,3)'];

% homogeneous points
x0hm = [x0m,ones(2,1)];    % (U0)
zhm  = [zm ,ones(2,1)];    % (V)

% ancillary vectors  
l2hm = x0hm * Em;          % (L)

% ancillary matrix
P0m = [eye(2)    ;[0 0]];  % (P0)

At  = zeros(6, 5);
        % Att = zeros(3, 5);

% determine variances of constraints
Bmt  = zeros(6,12);
for i = 1:2
%     Btt  = zeros(3, 6);
%     Btt(1:2,1:4) = kron(eye(2),l2hm(i,1:2));
%     Btt(1:2,5:6) = Em(1:2,1:2);
%     Btt(3,5:6)   = 2*l2hm(i,1:2);
%     
%     Bt =  [P0m,x0hm(i,:)']' * Em * kron([0,0,1],P0m)...
%         + kron(eye(3),l2hm(i,1:2));
%     
    Bt = [P0m,x0hm(i,:)']' * Em * [zeros(3,4),P0m] ...
        + [l2hm(i,1:2) zeros(1,4); ...
           0 0 l2hm(i,1:2) 0 0; ...
           zeros(1,4) l2hm(i,1:2)];
       
    Bmt(3*i-2:3*i,6*i-5:6*i)= Bt;
end

%nB = null(Bv');
[Q,~]=qr(Bv);
nB = Q(:,2:3);

Dm = zeros(6,5);
for i = 1:2
    x0h = x0hm(i,:)';       
    zh  = zhm(i,:)';        
    Xm  = [P0m,x0h];
    A0m = [Am(3*i-2:3*i-1,1:2);[0 0]];
    Zm  = [A0m,zh];
    % db_r
        %     Att(1:3,1:2) = ...
        %         [(-P0m'*calc_S(z1h) + A0m'*Rm*calc_S(x0h)) ;...
        %         2*(calc_S(z1h)*x0h)'] ...
        %         * nB;
    
    K1 = Xm'*calc_S(Rm'*zh)'*nB;  % (K1 * Jr)
    K3 = Zm'*Rm*calc_S(x0h)*nB;   % (K3 * Jr)
    At(3*i-2:3*i,1:2) = K1+K3;
    
    % dr
        %     Att(1:3,3:5) = ...
        %         [(P0m'*calc_S(Bv)*calc_S(z1h) + A0m'*Rm*calc_S(nh)');...
        %         2*(calc_S(nh)*z1h)'] ...
        %         * Rm';
    K2 =Xm'*Em*calc_S(zh);
    K4 =Zm'*calc_S(Em'*x0h)';
    At(3*i-2:3*i,3:5) = K2 + K4;
    
    Dm(3*i-2:3*i,:)   = At(3*i-2:3*i,:);
    
end
Bmt_5 = Bmt(indk,:);
Dm_5  = Dm(indk,:);

Caa = [Caam(1:6,:) zeros(6);zeros(6) Caam(7:12,:)];
Nm  = Dm_5' * inv(Bmt_5 * Caa * Bmt_5') * Dm_5;

CovM = inv(Nm);


% check Jy by variance propagation
% Wgg  = [BtCB(:,:,1),zeros(3);zeros(3) BtCB(:,:,2)];
% Di   = pinv(Dm);
% CovM_new = Di*Wgg*Di';
% 
% A1 = Am(1:2,1:3);
% A2 = Am(1:2,4:6);
% mv = [A1(:);...
%       A2(:)];
% pv = [x0m(:);
%       zm(:);
%       Bv;...
%       Rm(:);...
%       indk(:)];
% [~,~,Jy]=var_prop_classical(@c_from_y_5,zeros(12,1),...
%     [Caam(:,:,1),zeros(6);zeros(6),Caam(:,:,2)],pv);
% 
% % check Jtheta
% A1 = Am(1:2,1:3);
% A2 = Am(1:2,4:6);
% mv = [A1(:);...
%       A2(:)];
% pv = [A1(:);...
%       A2(:);...
%       x0m(:);
%       zm(:);
%       Bv;...
%       Rm(:);...
%       indk(:)];
% [~,~,Jx]=var_prop_classical(@c_from_theta_5,zeros(5,1),...
%     eye(5)*0.0000001,pv);
% 



end



