%% check precision of E from 2ACs
%
% 1. camera P=[I,0]
% generate 2. camera randomly R[I,-B]
% generate 2 random planes (X,N)
% generate 2 small offsets b
% generate 2 CovM (A: 0.1/s^2, b: 0.2/s)
% generate 2 Acs
% solve for E
% check for correctness of solution
% 
% wf 2019/12/12



clearvars
close all
clc


disp(' ')
disp(' ')
disp('          ========= Begin of Demo E from 2 Acs ===========')


sugr_INIT
ss = plot_init;

% random number intialization (if \= 0 then seed varies from call to call)
init_rand      = 0;
% init_rand = 210670;

initial_seed=init_rand_seed(init_rand);
display(['seed of random numbers ',num2str(initial_seed)])

% significance level
S = 0.99;
display(['significance level     ',num2str(S)])

%number of samples for checking covariance matrix
N_samples     = 1000;
%N_samples     = 200;
  N_samples     = 50;
%   N_samples     = 15;
%     N_samples     = 1;

% initiate results
est_bR_samples = zeros(N_samples,5);

% for fine correspondences: assume camera to ease specification
c          = 500;
pixel_size = 1/c;

% for camera pose
sigma_r = 0.5;              % stdv of rotation
sigma_B = 0.3;              % stdv of B = [1,0,0]+ Delta B

% for planes
sigma_X = 0.1;              % stdv of 3D point position
sigma_N = 0.1;              % stdv of normals

% for keypoints
sigma_bb = 0.5*pixel_size;             % stdv of bias of keypoints

% for least squares matching
scales   = [5,5];          % scales of points (pixels)
sigma_a0 = 0.1*pixel_size;             % normalized stdv of A
sigma_b0 = 0.3*pixel_size;             % normalized stdv of b




%% generate cameras
[Bv,Rm,E_true] = generate_camera_pair(sigma_r,sigma_B);

P1m  = [eye(3),zeros(3,1)];
P2m  = Rm * [eye(3),Bv];

% generate 2 planes,
[X1hev,Ahe1,X2hev,Ahe2] = generate_planes(sigma_X,sigma_N);

n1 = calc_S(Bv)*X1hev(1:3);
n2 = calc_S(Bv)*X2hev(1:3);
epipolar_angle_degree = atan2(norm(cross(n1,n2)),n1'*n2)*180/pi;
display(['epipolar angle         ',num2str(epipolar_angle_degree),''])



%% check direct estimation
disp('Monitor samples:')
iii = 0;
s_a   = zeros(1,N_samples);
s_ml  = zeros(1,N_samples);
start = cputime;
E_cputime = 0;
iii=0;
for ii = 1:N_samples
    monitor_samples(ii,N_samples);
    
    % generate two noisy Acs(x0,z0,A,Caa)
    [u0,v0,A,Caa,A_true] = generate_AC_pair(...
        Bv,Rm,X1hev,X2hev,Ahe1(1:2),Ahe2(1:2),...
        scales,sigma_bb,sigma_a0,sigma_b0);
    
    %% estimate E
    
    [E_est,Cm,Um,dv,Vm,indk] = ori_E_from_2ACs(u0,v0,A);
    E_cputime = E_cputime + toc;
    if ii==1
        disp(' ')
        display(['selected constraints    ',num2str(indk(:)')])
    end
    if size(E_est,1) > 0
        % next slution
        iii=iii+1;
        
        % recover parameters        
        %v  = v0 + [A(1:2,3,1)';A(1:2,3,2)'];
        v  = v0 + [A(1:2,3)';A(4:5,3)'];
        [t,R,~] = sugr_bR_from_E_uv (E_est,[u0,[1;1]],[v,[1;1]]);
        
        % determine differences
        est_bR_samples(iii,:) = [null(Bv')'*t;calc_v_from_S(R*Rm')];
    end
end

% AM = A_true;
% AM(:,1:3);
% AM(:,4:6);
% X0M = u0;
% ZM  = v0;

% determine theoretical covariance matrix using C
CovM_bR = CovM_bR_from_ACs_5(u0, v0, A_true, Caa, Bv, Rm, indk);

% % determine theoretical covariance matrix using C5
% CovM = CovM_br_from_2_ACs(Bv, Rm, Cm, Um, dv, Vm, Caa);
% 

% check theoretical covariance matrix
A1 = A_true(1:2,:);
A2 = A_true(4:5,:);
mv = [A1(:);...
      A2(:)];
pv = [u0(:);
      v0(:);
      Bv;...
      Rm(:)];
[bR_red,CovM_bR_num,~]=var_prop_classical(@E_from_2AC,mv,...
    [Caa(1:6,:),zeros(6);zeros(6),Caa(7:12,:)],pv);


Sigma_Rho_C=covariance_2_sigma_rho(CovM_bR)

sqrt_cond=sqrt(cond(CovM_bR))

%
rho_degree = 180/pi;
disp(strcat('sigma basis direction [degrees] = ',...
    num2str(rho_degree*sqrt(diag(CovM_bR(1:2,1:2)))')));
disp(strcat('sigma rotation        [degrees] = ',...
    num2str(rho_degree*sqrt(diag(CovM_bR(3:5,3:5)))')));


%% check algorithm
N_samples_red = iii;

if N_samples_red > 5
    
   rho_milli_rad = 1000;
    
    figure
    fz = 25;        % for labels
    set(gca,'fontsize',fz)
    
    s = rho_milli_rad*[est_bR_samples(:,1);est_bR_samples(:,2)];
    [~,bins]=hist(s,2*ceil(sqrt(N_samples_red)));
    
    subplot(1,5,1);
    hold on
    hist(rho_milli_rad*est_bR_samples(:,1),bins)
    set(gca,'fontsize',fz)
    %title(['\fontsize{25} b_x, \sigma = ',num2str(rho_milli_rad*stds(1)),'\mu rad']) 
    xlabel('bx','fontsize',fz)
    ylabel('#','fontsize',fz)
    
    
    subplot(1,5,2);
    hist(rho_milli_rad*est_bR_samples(:,2),bins)
    set(gca,'fontsize',fz)
    %title(['\fontsize{25} b_y, \sigma = ',num2str(rho_milli_rad*stds(2)),'\mu rad'])
    xlabel('by','fontsize',fz)
    ylabel('#','fontsize',fz)
    
    subplot(1,5,3);
    hist(rho_milli_rad*est_bR_samples(:,3),2*ceil(sqrt(N_samples_red)))
    set(gca,'fontsize',fz)
    %title(['\fontsize{25} r_1, \sigma = ',num2str(rho_milli_rad*stds(3)),'\mu rad'])
    xlabel('r1','fontsize',fz)
    ylabel('#','fontsize',fz)
    
    subplot(1,5,4);
    hist(rho_milli_rad*est_bR_samples(:,4),2*ceil(sqrt(N_samples_red)))
    set(gca,'fontsize',fz)
    %title(['\fontsize{25} r_2, \sigma = ',num2str(rho_milli_rad*stds(4)),'\mu rad'])
    xlabel('r2','fontsize',fz)
    ylabel('#','fontsize',fz)
    
    subplot(1,5,5);
    hist(rho_milli_rad*est_bR_samples(:,5),2*ceil(sqrt(N_samples_red)))
    set(gca,'fontsize',fz)
    %title(['\fontsize{25} r_3, \sigma = ',num2str(rho_milli_rad*stds(5)),'\mu rad'])  
    xlabel('r3','fontsize',fz)
    ylabel('#','fontsize',fz)
    
    
    % check of 2AC -> E
    check_estimation_result(0,zeros(5,1),CovM_bR,zeros(N_samples_red,1),...
        est_bR_samples(1:N_samples_red,:),S,' 2AC_E');
    
    %% plot emprirical and theoretical CovM
    imin = 0;
    imax = 0;
    rhomax = 0;
    for i=1:4
        for j=i+1:5
            if abs(Sigma_Rho_C(i,j)) > rhomax
                rhomax = Sigma_Rho_C;
                imin = i;
                imax = j;
            end
        end
    end     
    indc = [imin, imax];
    f=10^8;
    C_theor_2.mean = [0;0]; 
    C_theor_2.cov = CovM_bR(indc,indc)*f;
    CovM_bR_emp = cov(est_bR_samples);
    C_emp_2.mean = [0;0];
    C_emp_2.cov = CovM_bR_emp(indc,indc)*f;
    figure
    hold on
    set(gca,'fontsize',fz);
    plot_ellipse(C_emp_2,'-r');
    plot_ellipse(C_theor_2,'-b');
    s1 = sqrt(C_emp_2.cov(imin,imin));
    s2 = sqrt(C_emp_2.cov(imax,imax));
    sm = max(s1,s2);
    plot([0,0],1.2*[-sm,+sm],'--k','LineWidth',2);
    plot(1.2*[-sm,+sm],[0,0],'--k','LineWidth',2);
    axis equal
end


disp('          ========= End of Demo ===========')

