% ORI_E_2_AFFINE_POINT Berechnung der E-Matrizen (<=10) aus 2 affinen
% Korrespondenzen (Euklidisch)
% Eingabe:
%  u:      2x2 Matrix aus 2 Referenzpunkten u(i,:) in der 1. Kamera (x0)
%  v:      2x2 Matrix aus 2 Referenzpunkten v(i,:) in der 2. Kamera (z0)
%  A:      6x3 Affinitten [A1;A2]
% Ausgabe:
%  Es: Nx3x3 Matrix mit N<=10, die die mglichen E-Matrizen enthlt
%
% @author Christian Beder/Wolfgang Frstner
function [Es,cs,C,U,d,V,indk] = oris_2ACs_2_E(u,v,A)
    % built the DLT-Matrix
    % 	for i = 1:5
    %       DLT(i,:) = kron (u(i,:)', v(i,:)')';
    % 	end;
    for i=1:2
        uh = [u(i,:)';1];                  % x0
        %z  = v(i,:)' + A(1:2,3,i);         % z
        z   = v(i,:)' + A(3*i-2:3*i-1,3);
        zh = [z;1];
        U  = [eye(2)       u(i,:)';[0 0 1]];
        % V  = [A(1:2,1:2,i) z      ;[0,0,1]];
        V = [A(3*i-2:3*i-1,1:2) z; [0,0,1]];
        DLT(3*i-2:3*i,:) = kron(U',zh') + kron(uh',V');
    end
    
    % reduce DLT
	[Ud,~,~] = svd(DLT,'econ');
    cos = abs(Ud(:,6))./sqrt(diag(DLT*DLT'));
    %[~,k] = min(cos([1,2,4,5]));
    [~,k] = min(cos);
    indk = setdiff([1:6],k);
    DLTr = DLT(indk,:);  
	[Udr,Ddr,Vdr] = svd(DLTr);
	N = Vdr(:,6:9);

	% calculate the 4 base-vectors, that span the space of possible E-matrices 
	for i = 1:4 
      E(:,:,i) = [N(1:3,i),N(4:6,i),N(7:9,i)]';
	end;
	
	% calculate the conditions for the E-matrix
	A1 = zeros(9,4,4,4);
	for i = 1:4
      for j = 1:4
        for k = 1:4
          px = length(find([i,j,k]==1))+1;
          py = length(find([i,j,k]==2))+1;
          pz = length(find([i,j,k]==3))+1;
          condition2(:,:) = E(:,:,i) * E(:,:,j)' * E(:,:,k) - 1/2 * trace(E(:,:,i) * E(:,:,j)') * E(:,:,k);
          A1(1:9,px,py,pz) = A1(1:9,px,py,pz) + condition2(:);
        end;
      end;
	end;
	
	% fill in the coefficients into the matrix A2 as depicted in the paper
	A2 = zeros(10,20);
	A2(2:10,1)  = A1(:, 4,1,1);
	A2(2:10,2)  = A1(:, 1,4,1);
	A2(2:10,3)  = A1(:, 3,2,1);
	A2(2:10,4)  = A1(:, 2,3,1);
	A2(2:10,5)  = A1(:, 3,1,2);    
	A2(2:10,6)  = A1(:, 3,1,1);
	A2(2:10,7)  = A1(:, 1,3,2);    
	A2(2:10,8)  = A1(:, 1,3,1);
	A2(2:10,9)  = A1(:, 2,2,2);
	A2(2:10,10) = A1(:, 2,2,1);
	A2(2:10,11) = A1(:, 2,1,3);
	A2(2:10,12) = A1(:, 2,1,2);
	A2(2:10,13) = A1(:, 2,1,1);
	A2(2:10,14) = A1(:, 1,2,3);
	A2(2:10,15) = A1(:, 1,2,2);
	A2(2:10,16) = A1(:, 1,2,1);
	A2(2:10,17) = A1(:, 1,1,4);
	A2(2:10,18) = A1(:, 1,1,3);
	A2(2:10,19) = A1(:, 1,1,2);
	A2(2:10,20) = A1(:, 1,1,1);
	
	% fill in the determinant condition
    A2(1,:) = ori_multo2(ori_multo1(squeeze(E(1,2,:)), squeeze(E(2,3,:)))-ori_multo1(squeeze(E(1,3,:)), squeeze(E(2,2,:))),squeeze(E(3,1,:))) ...
         +ori_multo2(ori_multo1(squeeze(E(1,3,:)), squeeze(E(2,1,:)))-ori_multo1(squeeze(E(1,1,:)), squeeze(E(2,3,:))),squeeze(E(3,2,:))) ...
         +ori_multo2(ori_multo1(squeeze(E(1,1,:)), squeeze(E(2,2,:)))-ori_multo1(squeeze(E(1,2,:)), squeeze(E(2,1,:))),squeeze(E(3,3,:)));
	
    % Make it GrevLex (for Eigenvalue-solution, B-Matrix is then no longer
    % constructed correctly!!!)
    % TODO: Incorporate the GrevLex order into the above and multo1 and
    % multo2...
    A2 = A2(:, [1,3,5,4,9,11,2,7,14,17,  6,10,12,8,15,18,13,16,19,20]);
    
	% calculate the Gauss-Jordan-elimination
 	A = rref(A2);

    % Build action matrix
    A_f = [-A(1:6, 11:20); eye(3), zeros(3), zeros(3,1), zeros(3); zeros(1,3), zeros(1,3), 1, zeros(1,3)];    
    
    % get real Eigenvectors
    [V,D] = eig(A_f);
    realV = V(:,find (imag(diag(D))==0));
    
    Es = zeros(0, 3,3); 
    cs  = zeros(0);
    for i=1:size(realV,2)
        E_est = realV(7,i)*E(:,:,1) + realV(8,i)*E(:,:,2) + realV(9,i)*E(:,:,3) + realV(10,i)*E(:,:,4);
        % It is necessary to check the condition, if A2 and hence A is
        % singular... (ChB, 8.11.05)
        E_est = E_est / norm(E_est);
        if max(abs(E_est * E_est' * E_est - 1/2 * trace(E_est * E_est') * E_est))<1e-10
            Es(end+1,:,:) = E_est;
            Et            = E_est';
            cs(end+1)     = norm(DLT*Et(:));
        end;
    end;
    
    C = DLTr';
    U = Udr;
    d = diag(Ddr);
    V = Vdr(:,1:6);

return
